from functools import wraps
from typing import Text, Union

from django.apps import apps
from django.db.models import Model

LOCK_MODES = (
    "ACCESS SHARE",
    "ROW SHARE",
    "ROW EXCLUSIVE",
    "SHARE UPDATE EXCLUSIVE",
    "SHARE",
    "SHARE ROW EXCLUSIVE",
    "EXCLUSIVE",
    "ACCESS EXCLUSIVE",
)


def require_lock(model: Union[Text, Model], lock: Text):
    """
    Decorator for PostgreSQL's table-level lock functionality

    Example:
        @atomic
        @require_lock(MyModel, 'ACCESS EXCLUSIVE')
        def myview(request)
            ...

    PostgreSQL's LOCK Documentation:
    http://www.postgresql.org/docs/8.3/interactive/sql-lock.html
    """

    def require_lock_decorator(view_func):
        @wraps(view_func)
        def wrapper(*args, **kwargs):
            if lock not in LOCK_MODES:
                raise ValueError("%s is not a PostgreSQL supported lock mode.")
            from django.db import connection

            if not isinstance(model, Model):
                true_model = apps.get_model(model)
            else:
                true_model = model

            cursor = connection.cursor()
            # noinspection PyProtectedMember
            cursor.execute(
                "LOCK TABLE %s IN %s MODE" % (true_model._meta.db_table, lock)
            )
            return view_func(*args, **kwargs)

        return wrapper

    return require_lock_decorator



from django.core.files.images import get_image_dimensions
from django.forms import ModelForm
from django.utils.translation import ugettext_lazy as _


class PictureSizeForm(ModelForm):
    """
    Checks that picture fields listed in FIELDS have the right image size

    Notes
    -----
    This is here for compatibility reasons, I don't know why this wasn't made
    as a field initially.

    Examples
    --------

    >>> class PressReleaseForm(PictureSizeForm):
    >>>     FIELDS = {"thumbnail": [234, 330]}
    >>>
    >>>     class Meta:
    >>>         model = PressRelease
    >>>         exclude = []
    """

    FIELDS = {}

    def clean(self):
        data = super().clean()

        for field, expect in self.FIELDS.items():
            if not data.get(field):
                continue

            actual = get_image_dimensions(data[field])

            if tuple(expect) != tuple(actual):
                self.add_error(
                    field,
                    _(
                        "Vous avez envoyé une image de %(actual_size)s alors "
                        "qu'est attendue une image de %(expected_size)s."
                    )
                    % {
                        "actual_size": f"{actual[0]}x{actual[1]}px",
                        "expected_size": f"{expect[0]}x{expect[1]}px",
                    },
                )



from io import BytesIO
from multiprocessing import Pool

import zopfli
from django.contrib.staticfiles.storage import ManifestStaticFilesStorage
from django.contrib.staticfiles.utils import matches_patterns
from django.core.files.base import File
from psutil import cpu_count
from tqdm import tqdm


class GzipMixin:
    """
    Brings the Gzip-ability if mixed with a storage. Uses Zopfli for
    compression.

    Notes
    -----
    The base mixin was stolen from django-pipeline and then modified a lot
    https://django-pipeline.readthedocs.io/en/latest/storages.html
    """

    gzip_patterns = ("*.css", "*.js", "*.svg", "*.ttf")

    def _compress(self, original_file):
        c = zopfli.ZopfliCompressor(zopfli.ZOPFLI_FORMAT_GZIP)
        z = c.compress(original_file.read()) + c.flush()
        return File(BytesIO(z))

    def handle_file(self, path):
        """
        Handles a single path and generates the compressed file if required.
        It will create a .gz equivalent of that file.
        Parameters
        ----------
        path
            Path to the file to compress
        """

        if path:
            if not matches_patterns(path, self.gzip_patterns):
                return

            try:
                original_file = self.open(path, mode="rb")
            except FileNotFoundError:
                pass
            else:
                gzipped_path = "{0}.gz".format(path)

                if self.exists(gzipped_path):
                    self.delete(gzipped_path)

                gzipped_file = self._compress(original_file)
                gzipped_path = self.save(gzipped_path, gzipped_file)

                return gzipped_path, gzipped_path, True

    def post_process(self, paths, dry_run=False, **options):
        """
        For each path to post-process, potentially compress the files. That
        is done using a process pool that will parallelize the processing to
        speed it up.
        """

        super_class = super()

        if hasattr(super_class, "post_process"):
            for name, hashed_name, processed in super_class.post_process(
                paths.copy(), dry_run, **options
            ):
                if hashed_name != name:
                    paths[hashed_name] = (self, hashed_name)
                yield name, hashed_name, processed

        if dry_run:
            return

        cpus = (cpu_count(logical=False) or 1) + 1

        with Pool(processes=cpus) as pool:
            for out in tqdm(
                pool.imap_unordered(self.handle_file, paths),
                unit="file",
                total=len(paths),
            ):
                if out:
                    yield out


class GzipManifestStaticFilesStorage(GzipMixin, ManifestStaticFilesStorage):
    """
    Almost like the regular ManifestStaticFilesStorage, except it will create
    .gz files for all the text assets.
    """



from time import sleep
from typing import Callable

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.utils.timezone import now

ResponseGetter = Callable[[HttpRequest], HttpResponse]


class NowMiddleware:
    """
    This middleware adds a "now()" method to request objects, this way a
    common timestamp is available within the scope of the request. That's
    useful by example for token validity: if a token expires at a given time
    it allows to consider that the request conceptually happens instantly and
    every time you measure the token's validity you do so based on the
    request's global timestamp and not the actual time (which is just
    milliseconds away).
    """

    def __init__(self, get_response: ResponseGetter):
        """
        Storing away
        """

        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """
        Monkey-patching the request object with the custom now() function
        """

        fixed_now = now()
        object.__setattr__(request, "now", lambda: fixed_now)

        return self.get_response(request)


class SlowMiddleware:
    """
    That's a middleware that will simulate a latency on all requests.

    Notes
    -----
    You can set it by putting the number of seconds you want to sleep in the
    SLOW_MIDDLEWARE_LATENCY Django settings.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        to_sleep = getattr(settings, "SLOW_MIDDLEWARE_LATENCY", None)

        if to_sleep:
            sleep(to_sleep)

        return self.get_response(request)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['django_wools']

package_data = \
{'': ['*']}

install_requires = \
['django>1.7.0', 'psutil>2.0.0', 'tqdm>3.0.0', 'zopflipy>1.1']

setup_kwargs = {
    'name': 'django-wools',
    'version': '0.1.4',
    'description': 'Django tools from WITH',
    'long_description': 'Django Wools\n============\n\nDjango tools from WITH.\n\nThat\'s a collection of things that we at [WITH](https://with-madrid.com/) got\ntired of copy/pasting in every project.\n\n## Install\n\n```\npip install django_wools\n```\n\n## Included Wools\n\n### Storage\n\n#### `django_wools.storage.GzipManifestStaticFilesStorage`\n\nThat\'s a sub-class of the \n[ManifestStaticFilesStorage](https://docs.djangoproject.com/en/3.0/ref/contrib/staticfiles/#manifeststaticfilesstorage)\nbut that makes sure that along with all the files comes a `.gz` version which\nis easy to pick up for nginx (or other static files server).\n\n### Middlewares\n\n#### `django_wools.middlewares.NowMiddleware`\n\nSuppose that you have a content that is available up until a given date. When\nthe date is passed then everything related to this content expires. However,\nin order to do this, you\'re probably going to make several request, possibly in\nloosely connected parts of your code. In those cases, when looking at the time,\nthe clock will show different value as the time passes between calls. It means\nthat you could very well end up with one half of your code considering that the\nobject is still valid but the other half that it expired.\n\nIn order to prevent this, the simplest is to consider that the time is fixed\nand that the code executes instantly at the moment of the request. The goal\nof this middleware is to save the current time at each request and then to\nprovide an easy way to get the current time through the request.\n\nIf the middleware is activated, you should be able to get the time like this:\n\n```python\nfrom time import sleep\nfrom django.shortcuts import render\n\ndef my_view(request):\n    print(f"Now is {request.now()}")\n    sleep(42)\n    print(f"Now is still {request.now()}")\n\n    return render(request, "something.html", {"now": request.now()})\n```\n\n#### `django_wools.middlewares.SlowMiddleware`\n\nWhen developing a SPA or hybrid app, you want to make sure that the app is\nstructurally ready to handle load times from the server, even if the connection\nis a bit shaky. Also, you want to make sure that not too many requests are\nsent.\n\nIn order for you to fully realize how slow your website is going to be on a bad\nconnection, th e SlowMiddleware will automatically add a delay before replying\nto each request.\n\nAdd this to your `settings.py`\n\n```python\nMIDDLEWARE = [\n    # ...\n    "django_wools.middlewares.SlowMiddleware",\n]\n\nSLOW_MIDDLEWARE_LATENCY = 1 if DEBUG else 0\n```\n\nBy doing this, your requests will be added a 1s delay if the `DEBUG` mode is\nenabled.\n\n### Database\n\n#### `django_wools.db.require_lock`\n\nProvides a way to explicitly generate a PostgreSQL lock on a table.\n\nBy example:\n\n```python\nfrom django.db.transaction import atomic\nfrom django_wools.db import require_lock\n\nfrom my_app.models import MyModel\n\n\n@atomic\n@require_lock(MyModel, \'ACCESS EXCLUSIVE\')\ndef myview(request):\n    # do stuff here\n```\n',
    'author': 'Rémy Sanchez',
    'author_email': 'remy.sanchez@with-madrid.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/WithIO/django-wools',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
